@@ -60,9 +60,37 @@ an integer column named `id`, the ``belongs_to`` relationship will
 automatically add an integer column named `owner_id` to the entity, with a 
 foreign key referencing the `id` column of the `Person` entity.
 
-In addition to the keyword arguments inherited from SQLAlchemy, ``belongs_to``
-relationships accept an optional ``colname`` keyword argument, used to specify
-a custom name for the column which will be created.
+In addition to the keyword arguments inherited from SQLAlchemy's relation 
+function, ``belongs_to`` relationships accept the following optional arguments
+which will be directed to the created column:
+
++----------------------+------------------------------------------------------+
+| Option Name          | Description                                          |
++======================+======================================================+
+| ``colname``          | Specify a custom column name.                        |
++----------------------+------------------------------------------------------+
+| ``nullable``         | True if this column should allow nulls. Defaults to  |
+|                      | True unless this column is a primary key column.     |
++----------------------+------------------------------------------------------+
+| ``column_kwargs``    | A dictionary holding any other keyword argument you  |
+|                      | might want to pass to the Column.                    |
++----------------------+------------------------------------------------------+
+
+The following optional arguments are also supported to customize the
+ForeignKeyConstraint that is created:
+
++----------------------+------------------------------------------------------+
+| Option Name          | Description                                          |
++======================+======================================================+
+| ``use_alter``        | If True, SQLAlchemy will add the constraint in a     |
+|                      | second SQL statement (as opposed to within the       |
+|                      | create table statement). This permits to define      |
+|                      | tables with a circular foreign key dependency        |
+|                      | between them.                                        |
++----------------------+------------------------------------------------------+
+| ``constraint_kwargs``| A dictionary holding any other keyword argument you  |
+|                      | might want to pass to the Constraint.                |
++----------------------+------------------------------------------------------+
 
 `has_one`
 ---------
@@ -98,7 +126,7 @@ them being a `Person`. This could be expressed like so:
 
 Note that an ``has_many`` relationship **cannot exist** without a 
 corresponding ``belongs_to`` relationship in the other way. This is because the
-``has_one`` relationship needs the foreign key created by the ``belongs_to`` 
+``has_many`` relationship needs the foreign key created by the ``belongs_to`` 
 relationship.
 
 `has_and_belongs_to_many`
@@ -193,10 +221,10 @@ class Relationship(object):
         '''
         Sets up the relationship, creates foreign keys and secondary tables.
         '''
-        
+
         if not self.target:
             return False
-        
+
         if self.property:
             return True
 
@@ -206,19 +234,18 @@ class Relationship(object):
         
         return True
     
-    @property
     def target(self):
         if not self._target:
             path = self.of_kind.rsplit('.', 1)
             classname = path.pop()
-            
+
             if path:
                 # do we have a fully qualified entity name?
                 module = sys.modules[path.pop()]
             else: 
                 # if not, try the same module as the source
                 module = self.entity._descriptor.module
-            
+
             self._target = getattr(module, classname, None)
             if not self._target:
                 # This is ugly but we need it because the class which is
@@ -235,25 +262,31 @@ class Relationship(object):
                     return None
         
         return self._target
+    target = property(target)
     
-    @property
     def inverse(self):
-        #TODO: we should use a different value for when an inverse was searched
-        # for but none was found than when it hasn't been searched for yet so
-        # that we don't do the whole search again
         if not self._inverse:
             if self.inverse_name:
                 desc = self.target._descriptor
-                inverse = desc.relationships[self.inverse_name]
+                # we use all_relationships so that relationships from parent
+                # entities are included too
+                inverse = desc.all_relationships.get(self.inverse_name, None)
+                if inverse is None:
+                    raise Exception(
+                              "Couldn't find a relationship named '%s' in "
+                              "entity '%s' or its parent entities." 
+                              % (self.inverse_name, self.target.__name__)
+                          )
                 assert self.match_type_of(inverse)
             else:
                 inverse = self.target._descriptor.get_inverse_relation(self)
-        
+
             if inverse:
                 self._inverse = inverse
                 inverse._inverse = self
         
         return self._inverse
+    inverse = property(inverse)
     
     def match_type_of(self, other):
         t1, t2 = type(self), type(other)
@@ -283,6 +316,14 @@ class BelongsTo(Relationship):
     
     def __init__(self, entity, name, *args, **kwargs):
         self.colname = kwargs.pop('colname', None)
+        self.column_kwargs = kwargs.pop('column_kwargs', {})
+        if 'required' in kwargs:
+            self.column_kwargs['nullable'] = not kwargs.pop('required')
+
+        self.constraint_kwargs = kwargs.pop('constraint_kwargs', {})
+        if 'use_alter' in kwargs:
+            self.contraint_kwargs['use_alter'] = kwargs.pop('use_alter')
+        
         if self.colname and not isinstance(self.colname, list):
             self.colname = [self.colname]
         super(BelongsTo, self).__init__(entity, name, *args, **kwargs)
@@ -356,7 +397,8 @@ class BelongsTo(Relationship):
 
                 # we use a Field here instead of using a Column directly 
                 # because of add_field 
-                field = Field(pk_col.type, colname=colname, index=True)
+                field = Field(pk_col.type, colname=colname, index=True, 
+                              **self.column_kwargs)
                 source_desc.add_field(field)
 
                 self.foreign_key.append(field.column)
@@ -366,23 +408,25 @@ class BelongsTo(Relationship):
                 fk_colnames.append(colname)
 
                 # build the list of columns the foreign key will point to
-                fk_refcols.append(target_desc.tablename + '.' + pk_col.name)
+                fk_refcols.append("%s.%s" % (target_desc.entity.table.name,
+                                             pk_col.name))
 
                 # build up the primary join. This is needed when you have 
                 # several belongs_to relations between two objects
                 self.primaryjoin_clauses.append(field.column == pk_col)
             
-            # TODO: better constraint-naming?
-            #CHECKME: do we really need use_alter systematically?
+            # In some databases (at lease MySQL) the constraint name needs to 
+            # be unique for the whole database, instead of per table.
+            fk_name = "%s_%s_fk" % (self.entity.table.name, self.name)
             source_desc.add_constraint(ForeignKeyConstraint(
                                             fk_colnames, fk_refcols,
-                                            name=self.name +'_fk',
-                                            use_alter=True))
+                                            name=fk_name,
+                                            **self.constraint_kwargs))
     
     def create_properties(self):
         kwargs = self.kwargs
         
-        if self.entity is self.target:
+        if self.entity.table is self.target.table:
             if self.entity._descriptor.autoload:
                 cols = [col for col in self.target.table.primary_key.columns]
             else:
@@ -400,8 +444,19 @@ class HasOne(Relationship):
     uselist = False
 
     def create_keys(self):
-        # make sure the inverse is set up because it creates the
-        # foreign key we'll need
+        # make sure the inverse exists
+        if self.inverse is None:
+            raise Exception(
+                      "Couldn't find any relationship in '%s' which "
+                      "match as inverse of the '%s' relationship "
+                      "defined in the '%s' entity. If you are using "
+                      "inheritance you "
+                      "might need to specify inverse relationships "
+                      "manually by using the inverse keyword."
+                      % (self.target.__name__, self.name,
+                         self.entity.__name__)
+                  )
+        # make sure it is set up because it creates the foreign key we'll need
         self.inverse.setup()
     
     def create_properties(self):
@@ -412,9 +467,8 @@ class HasOne(Relationship):
         # are indeed useful, or remove them. It might be they are indeed
         # useless because of the primaryjoin, and that the remote_side is
         # already setup in the other way (belongs_to).
-        if self.entity is self.target:
+        if self.entity.table is self.target.table:
             kwargs['remote_side'] = self.inverse.foreign_key
-            print "kwargs", self.name, kwargs
         
         kwargs['primaryjoin'] = and_(*self.inverse.primaryjoin_clauses)
         kwargs['uselist'] = self.uselist
@@ -453,14 +507,58 @@ class HasAndBelongsToMany(Relationship):
             e1_desc = self.entity._descriptor
             e2_desc = self.target._descriptor
             
+            if e1_desc.autoload:
+                if not self.user_tablename:
+                    raise Exception(
+                        "Entity '%s' is autoloaded but relation '%s' has no "
+                        "secondary table name specified. You should specify "
+                        "it by using the tablename keyword."
+                        % (self.entity.__name__, self.name)
+                    )
+
+            # We use the name of the relation for the first entity 
+            # (instead of the name of its primary key), so that we can 
+            # have two many-to-many relations between the same objects 
+            # without having a table name collision. 
+            source_part = "%s_%s" % (e1_desc.tablename, self.name)
+
+            # And we use the name of the primary key for the second entity
+            # when there is no inverse, so that a many-to-many relation 
+            # can be defined without an inverse.
+            if self.inverse:
+                e2_name = self.inverse.name
+            else:
+                e2_name = '_'.join([key.column.name for key in
+                                    e2_desc.primary_keys])
+            target_part = "%s_%s" % (e2_desc.tablename, e2_name)
+            
+            if self.user_tablename:
+                tablename = self.user_tablename
+            else:
+                # we need to keep the table name consistent (independant of 
+                # whether this relation or its inverse is setup first)
+                if self.inverse and e1_desc.tablename < e2_desc.tablename:
+                    tablename = "%s__%s" % (target_part, source_part)
+                else:
+                    tablename = "%s__%s" % (source_part, target_part)
+
+            # In some databases (at lease MySQL) the constraint names need 
+            # to be unique for the whole database, instead of per table.
+            source_fk_name = "%s_fk" % source_part
+            if self.inverse:
+                target_fk_name = "%s_fk" % target_part
+            else:
+                target_fk_name = "%s_inverse_fk" % source_part
+
             columns = list()
             constraints = list()
 
             self.primaryjoin_clauses = list()
             self.secondaryjoin_clauses = list()
 
-            for num, desc, join_name in (('1', e1_desc, 'primary'), 
-                                         ('2', e2_desc, 'secondary')):
+            for num, desc, join_name, fk_name in (
+                    ('1', e1_desc, 'primary', source_fk_name), 
+                    ('2', e2_desc, 'secondary', target_fk_name)):
                 fk_colnames = list()
                 fk_refcols = list()
             
@@ -489,40 +587,13 @@ class HasAndBelongsToMany(Relationship):
                     join_list = getattr(self, join_name+'join_clauses')
                     join_list.append(col == pk_col)
                 
-                # TODO: better constraint-naming?
-                #CHECKME: do we really need use_alter systematically?
                 constraints.append(
                     ForeignKeyConstraint(fk_colnames, fk_refcols,
-                                         name=desc.tablename + '_fk', 
-                                         use_alter=True))
+                                         name=fk_name))
 
-            if self.user_tablename:
-                tablename = self.user_tablename
-            else:
-                # We use the name of the relation for the first entity 
-                # (instead of the name of its primary key), so that we can 
-                # have two many-to-many relations between the same objects 
-                # without having a table name collision. 
-                source_part = "%s_%s" % (e1_desc.tablename, self.name)
-
-                # And we use the name of the primary key for the second entity
-                # when there is no inverse, so that a many-to-many relation 
-                # can be defined without an inverse.
-                if self.inverse:
-                    e2_name = self.inverse.name
-                else:
-                    e2_name = '_'.join([key.column.name for key in
-                                        e2_desc.primary_keys])
-                target_part = "%s_%s" % (e2_desc.tablename, e2_name)
-
-                # we need to keep the table name consistent (independant of 
-                # whether this relation or its inverse is setup first)
-                if self.inverse and e1_desc.tablename < e2_desc.tablename:
-                    tablename = "%s__%s" % (target_part, source_part)
-                else:
-                    tablename = "%s__%s" % (source_part, target_part)
 
             args = columns + constraints
+            
             self.secondary_table = Table(tablename, e1_desc.metadata, *args)
     
     def create_properties(self):
@@ -545,6 +616,7 @@ class HasAndBelongsToMany(Relationship):
                (self.user_tablename == other.user_tablename or 
                 (not self.user_tablename and not other.user_tablename))
 
+
 belongs_to              = Statement(BelongsTo)
 has_one                 = Statement(HasOne)
 has_many                = Statement(HasMany)
